var _ = require('underscore');
var helpers = require('./util/helpers');
var car = helpers.car;
var cdr = helpers.cdr;

// forward declarations
var loopEval;

var loopEvalCollection = function(array) {
  return _.map(array, function(node) {
    return loopEval(node);
  });
};

var loopEval = function(syntaxTree) {
  var contents = syntaxTree.contents;

  switch (syntaxTree.type) {
    case 'list':
      if (contents.length === 0) {
        throw new Error("list of length 0!");
      }

      var firstToken = car(contents);
      var remainingTokens = cdr(contents);

      if (firstToken.type === 'list') {
        return {
          type: 'funcall',
          function: loopEval(car(contents)),
          arguments: {
            type: 'list',
            contents: loopEvalCollection(cdr(contents))
          }
        };
      }

      var body;

      if (firstToken.type === 'id') {
        if (firstToken.contents === 'lambda' || firstToken.contents === 'function') {
          var formalArgs = car(remainingTokens);
          body = loopEvalCollection(cdr(remainingTokens));

          return {
            type: 'funcall',
            function: firstToken,
            arguments: {
              type: 'list',
              contents: [
                formalArgs,
                {
                  type: 'list',
                  contents: body
                }
              ]
            }
          };
        } else { // user defined or built in function
          body = loopEvalCollection(remainingTokens);

          return {
            type: 'funcall',
            function: firstToken,
            arguments: {
              type: 'list',
              contents: body
            }
          };
        }
      }

      if (firstToken.type === 'prop-access') {
        body = loopEvalCollection(remainingTokens);

        return {
          type: 'funcall',
          function: loopEval(firstToken),
          arguments: {
            type: 'list',
            contents: body
          }
        };
      }

      var errorMessage = "Parse error near: `" + firstToken.contents + "`";
      if (firstToken.sourceInfo) {
        errorMessage += [
          ' ',
          '-- line: ',
          firstToken.sourceInfo.first_line,
          " -- col: ",
          firstToken.sourceInfo.first_column,
          "-",
          firstToken.sourceInfo.last_column
        ].join('');
      }

      throw new Error(errorMessage);
    case 'prop-access':
      var key = contents.key;
      var value = contents.value;

      return {
        type: 'prop-access',
        key: loopEval(key),
        value: loopEval(value)
      };
    case 'id':
    case 'string':
    case 'number':
    case 'comment':
      return syntaxTree;
    default:
      throw new Error('unknown type of syntaxTree, type: ' + syntaxTree.type);
  }
};

module.exports = function(branches) {
  return _.map(branches, function(branch) {
    return loopEval(branch);
  });
};